import asyncio
import os
from typing import Dict, List

from pyparsing import Optional
from pixivcat import BaseModel, MultiMediaBaseModel, BaseClient
from .user import Artist
from .session import main_loop, Session, PROXY
from aiohttp.client_exceptions import ClientConnectorError, ServerDisconnectedError, ClientPayloadError, ClientOSError


class Illustration(BaseModel):
    id: int
    title: str
    caption: str
    create_date: str
    height: int
    image_urls: Dict
    is_bookmarked: bool
    is_muted: bool
    meta_pages: List
    meta_single_page: Dict
    page_count: int
    restrict: int
    sanity_level: int
    series: Dict
    tags: List
    tools: List
    total_bookmarks: int
    total_comments: int
    total_view: int
    type: str
    user: Artist
    visible: bool
    width: int
    x_restrict: int

    async def _download(self, url: str, _fn: str, _path: str = None, _session=None) -> None:
        image_type = url.split(".")[-1]
        _fn = _fn if _fn else url.split("/")[-1]
        if not _fn.endswith(image_type):
            _fn = _fn[:-4]+image_type
        _fn = _fn if not _path else _path+_fn
        print(f"{_fn} downloading...")
        while 1:
            try:
                async with _session.get(url, headers={"Referer": "https://app-api.pixiv.net/"}) as resp:
                    if resp.status not in (200, 301):
                        print(f"{_fn} download fail, status code: {resp.status}")
                    data = await resp.read()
                    with open(_fn, "wb") as f:
                        f.write(data)
                    await asyncio.sleep(.1)
            except (
                ClientConnectorError,
                ServerDisconnectedError,
                ClientPayloadError
            ) as e:
                continue
            except Exception as e:
                print(f"download {_fn} error:{str(e)}")
                raise
            else:
                break
        print(f"{_fn} download completed")

    async def download_origin(self, path: str = None, filename: str = None) -> None:
        if not self.meta_pages and not self.meta_single_page:
            return
        if path and not os.path.exists(path):
            os.mkdir(path)
        if path and not path.endswith("/"):
            path += "/"
        origin_urls: List[str] = [self.meta_single_page.get("original_image_url"), ] if self.meta_single_page else [
            url.get("image_urls").get("original") for url in self.meta_pages]
        while 1:
            try:
                async with Session(proxy=PROXY.get()) as session:
                    for url in origin_urls:
                        await self._download(url, filename, path, session)
                        if filename:
                            break
            except ClientOSError as e:
                continue
            return

    async def download_medium(self, medium: str = "medium", filename: str = None, path: str = None):
        """download illust medium
            Args:
                medium(str): square_medium, medium(Default), large
        """
        if not self.image_urls:
            return
        if path and not os.path.exists(path):
            os.mkdir(path)
        if path and not path.endswith("/"):
            path += "/"
        medium_urls: List[str] = [urls.get("image_urls").get(
            medium) for urls in self.meta_pages] if self.meta_pages else [self.image_urls.get(medium), ]
        async with Session(proxy=PROXY.get()) as session:
            for url in medium_urls:
                await self._download(url, filename, path, session)
                if filename:
                    break


class Illustrations(MultiMediaBaseModel):
    def __init__(self, client: "BaseClient", resp: Dict) -> None:
        self.illusts = [Illustration(**illust)
                        for illust in resp.pop("illusts")]
        super().__init__(client=client, **resp)

    def download_origins(self, path: str):
        try:
            main_loop = asyncio.get_running_loop()
        except RuntimeError:
            print("warnning: cannot get running loop")
            main_loop = main_loop.get()
        if not self.illusts:
            return
        for illust in self.illusts:
            main_loop.create_task(illust.download_origin(path=path))

    def download_mediums(self, path: str, medium: str = "medium"):
        """download illust medium
            Args:
                medium(str): square_medium, medium(Default), large
        """
        try:
            main_loop = asyncio.get_running_loop()
        except RuntimeError:
            print("warnning: cannot get running loop")
            main_loop = main_loop.get()
        if not self.illusts:
            return
        for illust in self.illusts:
            main_loop.create_task(
                illust.download_medium(path=path, medium=medium))



from pixivcat.user import Artists
from pixivcat.comments import Comments
from pixivcat.client import Client
from typing import Dict, List, Optional
from asyncio import AbstractEventLoop
from pixivcat.illust import Illustration, Illustrations
from pixivcat.novel import Novel, NovelSeries, Novels
from pixivcat.previews import Previews
from pixivcat.user import Artist, Profile, ProfilePublicity, Workspace
from pixivcat.bookmark import BookmarkDetail, Bookmarks
from pixivcat.tag import Tag


class AppClient(Client):
    url: str = "https://app-api.pixiv.net"

    def __init__(self, refresh_token: str = None, proxy: str = None, url: str = url, loop: Optional[AbstractEventLoop] = None) -> None:
        super().__init__(refresh_token, proxy, url, loop)

    async def user_detail(self, user_id: int, **kws) -> Artist:
        """Get user information by user ID"""
        response = await self.request("GET", f"/v1/user/detail?user_id={user_id}")
        artist = Artist(**response.get("user"))
        artist.profile = Profile(**response.get("profile"))
        artist.profile_publicity = ProfilePublicity(
            **response.get("profile_publicity"))
        artist.workspace = Workspace(**response.get("workspace"))
        return artist

    async def user_illusts(self, user_id: int, type: str = "illust") -> Illustrations:
        """Get user illustrations by user ID"""
        response = await self.request("GET", f"/v1/user/illusts?user_id={user_id}&type={type}")
        return Illustrations(self, response)

    async def user_bookmarks_illust(self, user_id: int, restrict="public", max_bookmark_id=None, tag=None) -> Illustrations:
        """Get user bookmarks by user ID"""
        params = f"?user_id={user_id}&restrict={restrict}"
        params += f"&max_bookmark_id={max_bookmark_id}" if max_bookmark_id else ""
        params += f"&tag={tag}" if tag else ""
        response = await self.request("GET", f"/v1/user/bookmarks/illust{params}")
        return Illustrations(self, response)

    async def user_related(self, seed_user_id: int, offset: int = 0) -> Previews:
        """Get user related previews by ...by user Id"""
        response = await self.request("GET", f"/v1/user/related?seed_user_id={seed_user_id}&offset={offset}")
        return Previews(self, response)

    async def illust_follow(self, restrict: str = "public", offset: int = 0) -> Illustrations:
        """Get to follow the illustrations of the artist"""
        response = await self.request("GET", f"/v2/illust/follow?restrict={restrict}&offset={offset}")
        return Illustrations(self, response)

    async def illust_detail(self, illust_id) -> Illustration:
        """get illust detail"""
        response = await self.request("GET", f"/v1/illust/detail?illust_id={illust_id}")
        return Illustration(**response.get("illust"))

    async def illust_comments(self, illust_id, offset: int = 0, include_total_comments: bool = True) -> Comments:
        """get illust comments"""
        response = await self.request("GET", f"/v1/illust/comments?illust_id={illust_id}&offset={offset}&includ_total_comments={str(include_total_comments).lower()}")
        return Comments(self, **response)

    async def illust_related(self, illust_id, offset: int = 0, seed_illust_ids: list = None) -> Illustrations:
        """Get user related previews by illust Id"""
        response = await self.request("GET", f"/v2/illust/related?illust_id={illust_id}&offset={offset}{'&seed_illust_ids='+str(seed_illust_ids) if seed_illust_ids else ''}")
        return Illustrations(self, response)

    async def illust_recommended(self,
                                 content_type="illust",
                                 offset: int = 0,
                                 include_ranking_label: bool = True,
                                 max_bookmark_id_for_recommend: int = None,
                                 min_bookmark_id_for_recent_illust: int = None,
                                 include_ranking_illusts: bool = None,
                                 bookmark_illust_ids: List[str] = None,
                                 include_privacy_policy: str = None
                                 ) -> Illustrations:
        """get illust recommend"""
        params = f"?content_type={content_type}&include_ranking_label={str(include_ranking_label).lower()}&offset={offset}"
        params += f"&max_bookmark_id_for_recommend={max_bookmark_id_for_recommend}" if max_bookmark_id_for_recommend else ''
        params += f"&min_bookmark_id_for_recent_illust={min_bookmark_id_for_recent_illust}" if min_bookmark_id_for_recent_illust else ''
        params += f"&include_ranking_illusts={str(include_ranking_illusts).lower()}" if include_ranking_illusts else ''
        params += f"&bookmark_illust_ids={','.join(bookmark_illust_ids)}" if bookmark_illust_ids else ''
        params += f"&include_privacy_policy={include_privacy_policy}" if include_privacy_policy else ''
        response = await self.request("GET", "/v1/illust/recommended"+params)
        return Illustrations(self, response)

    async def illust_ranking(self, mode: str = 'day', offset: int = 0, date: str = None) -> Illustrations:
        """get illusts rank

        Args:
            mode(str): day, week, month, day_male, day_female, week_original, week_rookie, day_r18, day_male_r18, day_female_r18, week_r18, week_r18g
            date(str): eg:2022-01-01

        """
        response = await self.request("GET", f"/v1/illust/ranking?mode={mode}&offset={offset}{f'&date={date}' if date else ''}")
        return Illustrations(self, response)

    async def trending_tags_illust(self) -> List[Tag]:
        """get trending tags"""
        response = await self.request("GET", "/v1/trending-tags/illust")
        return [Tag(**tag) for tag in response.get("trend_tags")]

    async def search_illust(self,
                            word: str,
                            search_target: str = 'partial_match_for_tags',
                            sort: str = 'date_desc',
                            offset: int = 0,
                            duration: str = "within_last_day",
                            start_date: str = None,
                            end_date: str = None) -> Illustrations:
        """search illust by word

            Args:

                word(str): search keyword

                search_target(str): 

                    partial_match_for_tags -> partial match tags
                    exact_match_for_tags -> full match tags
                    title_and_caption -> title match

                sort(str): date_desc, date_asc, popular_desc

                duration(str): within_last_day, within_last_week, within_last_month

                start_date/end_date (str):... you know...be like "2022-01-01"
        """
        params = f"?word={word}&search_target={search_target}&sort={sort}&offset={offset}"
        params += f"duration={duration}" if duration else ""
        params += f"start_date={start_date}" if start_date else ""
        params += f"end_date={end_date}" if end_date else ""
        response = await self.request("GET", "/v1/search/illust"+params)
        return Illustrations(self, response)

    async def search_novel(self,
                           word: str,
                           search_target: str = 'partial_match_for_tags',
                           sort: str = 'date_desc',
                           offset: int = 0,
                           merge_plain_keyword_results: bool = True,
                           include_translated_tag_results: bool = True,
                           start_date=None,
                           end_date=None,
                           ) -> Novels:
        """search novel by word

            Args:

                word(str): search keyword

                search_target(str): 

                    partial_match_for_tags -> partial match tags
                    exact_match_for_tags -> full match tags
                    keyword -> keyword match
                    text -> text match

                sort(str): date_desc, date_asc, popular_desc

                start_date/end_date (str):... you know...be like "2022-01-01"

        """
        params = f"?word={word}&search_target={search_target}&sort={sort}&offset={offset}"
        params += f"&merge_plain_keyword_results={str(merge_plain_keyword_results).lower()}"
        params += f"&include_translated_tag_results={str(include_translated_tag_results).lower()}"
        params += f"start_date={start_date}" if start_date else ""
        params += f"end_date={end_date}" if end_date else ""
        response = await self.request("GET", "/v1/search/novel"+params)
        return Novels(self, response)

    async def search_user(
        self,
        word: str,
        offset: int = 0,
        sort: str = 'date_desc',
        duration: str = "within_last_day"
    ):
        """search user by word

            Args:

                word(str): search keyword

                sort(str): date_desc, date_asc, popular_desc

                duration(str): within_last_day, within_last_week, within_last_month
        """
        params = f"?word={word}&offset={offset}&sort={sort}&duration={duration}"
        response = await self.request("GET", "/v1/search/user"+params)
        return Previews(self, response)

    async def illust_bookmark_detail(self, illust_id: int) -> BookmarkDetail:
        response = await self.request("GET", f"/v2/illust/bookmark/detail?illust_id={illust_id}")
        return BookmarkDetail(**response.get("bookmark_detail"))

    async def illust_bookmark_add(self, illust_id: int, restrict: str = "public", tags: List[str] = None) -> None:
        """add bookmark by illust id"""
        data = {
            "illust_id": illust_id,
            "restrict": restrict
        }
        if tags:
            data.update({"tags[]": ",".join(tags)})
        await self.request("POST", f"/v2/illust/bookmark/add", data=data)

    async def illust_bookmark_delete(self, illust_id: int) -> None:
        """delete bookmark by illust id"""
        await self.request("POST", "/v1/illust/bookmark/delete", data={"illust_id": illust_id})

    async def user_follow_add(self, user_id: int, restrict: int = "public") -> None:
        """follow user by user_id"""
        await self.request("POST", "/v1/user/follow/add", data={"user_id": user_id, "restrict": restrict})

    async def user_follow_delete(self, user_id: int) -> None:
        """unfollow user by user id"""
        await self.request("POST", "/v1/user/follow/delete", data={"user_id": user_id})

    async def user_bookmark_tags_illust(self,  restrict: str = 'public', offset: int = 0) -> Bookmarks:
        """get user bookmark tags"""
        response = await self.request("GET", f"/v1/user/bookmark-tags/illust?restrict={restrict}&offset={offset}")
        return Bookmarks(self, **response)

    # TODO
    async def user_following(self, user_id: int, restrict='public', offset: int = 0) -> Previews:
        """Get user following"""
        response = await self.request("GET", f"/v1/user/following?user_id={user_id}&restrict={restrict}&offset={offset}")
        return Previews(self, response)

    async def user_follower(self, user_id: int, offset: int = 0) -> Previews:
        """get user flowers by user_id"""
        response = await self.request("GET", f"/v1/user/follower?user_id={user_id}&offset={offset}")
        return Previews(self, response)

    async def user_mypixiv(self, user_id: int, offset: int = 0) -> Previews:
        """Get my pixiv friend by user id"""
        response = await self.request("GET", f"/v1/user/mypixiv?user_id={user_id}&offset={offset}")
        return Previews(self, response)

    async def user_blacklist(self, user_id: int, offset: int = 0) -> Artists:
        """get black list by user id"""
        response = await self.request("GET", f"/v2/user/list?user_id={user_id}&offset={offset}")
        return Artists(self, response)

    async def ugoira_metadata(self, illust_id: int):
        """Get ugoira metadata by illust_id"""
        response = await self.request("GET", f"/v1/ugoira/metadata?illust_id={illust_id}")
        return response

    async def user_novels(self, user_id: int, offset: int = 0) -> Novels:
        """Get user novels by user id"""
        response = await self.request("GET", f"/v1/user/novels?user_id={user_id}&offset={offset}")
        return Novels(self, response)

    async def novel_series(self, series_id: int, last_order: str = None) -> NovelSeries:
        """Get novels series detail"""
        response = await self.request("GET", f"/v2/novel/series?series_id={series_id}{f'&last_order={last_order}' if last_order else ''}")
        return NovelSeries(self, response)

    async def novel_detail(self, novel_id: int) -> Novel:
        """Get novel detail with novel id"""
        response = await self.request("GET", f"/v2/novel/detail?novel_id={novel_id}")
        return Novel(**response.get("novel"))

    async def novel_text(self, novel_id: int) -> Dict:
        response = await self.request("GET", f"/v1/novel/text?novel_id={novel_id}")
        response["series_prev"] = Novel(**response.pop("series_prev"))
        response["series_next"] = Novel(**response.pop("series_next"))
        return response

    async def illust_new(self, content_type: str = "illust", max_illust_id=None) -> Illustrations:
        """Get new illust"""
        response = await self.request("GET", f"/v1/illust/new?content_type={content_type}{f'&max_illust_id={max_illust_id}' if max_illust_id else ''}")
        return Illustrations(self, response)

    async def article(self, article_id: int):
        """Get articie"""
        headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/63.0.3239.132 Safari/537.36",
            "Referer": "https://www.pixiv.net"
        }
        response = self.request(
            "GET", f"https://www.pixiv.net/ajax/showcase/article?article_id={article_id}", headers=headers)
        return response



from asyncio import AbstractEventLoop
import asyncio
import hashlib

from pixivcat import BaseClient
from .user import User
from .session import Session, session, main_loop, PROXY
from typing import Any, Dict, NoReturn, Optional, Union


class URL:
    url: str

    def __init__(self, url=None) -> None:
        self.url = url

    def __call__(self, path: str, *args: Any, **kwds: Any) -> Any:
        if not path.startswith("/"):
            path += "/"
        return self.url + path


class Client(BaseClient):
    def __init__(self,
                 refresh_token: str = None,
                 proxy: str = None,
                 url: str = None,
                 loop: Optional[AbstractEventLoop] = None
                 ) -> None:
        self._url = URL(url=url)
        self._loop = loop or asyncio.new_event_loop()
        self._session = Session(proxy=proxy, loop=loop)
        session.set(self._session)
        main_loop.set(self._loop)
        self.refresh_token = refresh_token

        self.CLIENT_ID = "MOBrBDS8blbauoSck0ZfDbtuzpyT"
        self.CLIENT_SECRET = "lsACyCD94FhDUtGTXi3QzcFE2uU1hqtDaKeqrdwj"
        self.encryption = "28c1fdd170a5204386cb1313c7077b34f83e4aaf4aa829ce78c231e05b0bae2c"

        self.access_token = None
        self.user = None

    @classmethod
    def encrypt(self, data: str, encryption: str = "") -> None:
        mix = data+encryption
        return hashlib.md5(mix.encode("UTF-8")).hexdigest()

    def set_url(self, url: str) -> None:
        self._url = url

    async def request(self, method: str, url: str, **kwds) -> Dict:
        url = self._url(url) if url.startswith("/") else url
        async with self._session.request(method=method, url=url, **kwds) as resp:
            return await self.raise_status_code(resp)

    async def raise_status_code(self, response) -> Union[NoReturn, Dict]:
        if response.status not in (200, 301):
            raise Exception(f"request error, status code {(await response.json())}")
        return await response.json()

    async def auth(self, refresh_token: Optional[str] = None, headers: Dict = None):
        """login, use refresh_token"""
        data = {}
        headers = headers or {}
        if not headers.get("user-agent"):
            headers["user-agent"] = "PixivAndroidApp/5.0.234 (Android 11; Pixel 5)"
        self._session.update_headers(headers)
        data["grant_type"] = "refresh_token"
        data["client_id"] = self.CLIENT_ID
        data["client_secret"] = self.CLIENT_SECRET
        if not refresh_token and not self.refresh_token:
            raise ValueError("no refresh_token")
        data["refresh_token"] = refresh_token or self.refresh_token
        response = await self.request("POST", "https://oauth.secure.pixiv.net/auth/token", data=data, headers=headers)
        if not response:
            # raise no response exception
            raise Exception(f"get auth response error")
        self.access_token = response.get("access_token")
        self._session.update_headers(
            Authorization=f"Bearer {self.access_token}")
        self.user = User(**response.get("user"))

    async def __aenter__(self):
        if self.refresh_token:
            await self.auth()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """wait download complite"""
        while 1:
            tasks = asyncio.all_tasks()
            if len(tasks) <= 1:
                break
            await asyncio.sleep(1)
        if not self._session.closed:
            await self._session.close()
        if not self._loop.is_running():
            self._loop.close()



from typing import List
from pixivcat import BaseModel
from .client import BaseClient


class Comments(BaseModel):
    total_comments: int
    commenets: List
    next_url: str
    comment_access_control: int

    def __init__(self, client: BaseClient, **kws) -> None:
        self.client = client
        super().__init__(**kws)

    async def next(self) -> "Comments":
        response = await self.client.request("GET", self.next_url)
        return Comments(client=self.client, **response)



from pixivcat import BaseModel, MultiMediaBaseModel, BaseClient
from typing import Dict, List


class User(BaseModel):
    account: str = None
    id: str = None
    is_mail_authorized: bool = None
    is_premium: bool = None
    mail_address: str = None
    name: str = None
    profile_image_urls: Dict = None
    x_restrict: int = None
    sex: str = None


class ProfilePublicity(BaseModel):
    birth_day: str
    birth_year: str
    gender: str
    job: str
    pawoo: bool
    region: str


class Profile(BaseModel):
    address_id: int = None
    background_image_url: str = None
    birth: str = None
    birth_day: str = None
    birth_year: int = None
    country_code: str = None
    gender: str = None
    is_premium: bool = None
    is_using_custom_profile_image: bool = None
    job: str = None
    job_id: int = None
    pawoo_url: str = None
    region: str = None
    total_follow_users: int = None
    total_illust_bookmarks_public: int = None
    total_illust_series: int = None
    total_illusts: int = None
    total_manga: int = None
    total_mypixiv_users: int = None
    total_novel_series: int = None
    total_novels: int = None
    twitter_account: str = None
    twitter_url: str = None
    webpage: str = None


class Workspace(BaseModel):
    chair: str
    comment: str
    desk: str
    desktop: str
    monitor: str
    mouse: str
    music: str
    pc: str
    printer: str
    scanner: str
    tablet: str
    tool: str
    workspace_image_url: str


class Artist(BaseModel):
    id: int
    name: str
    account: str
    profile_image_urls: Dict
    profile: Profile = None
    workspace: Workspace = None
    profile_publicity: ProfilePublicity = None
    comment: str = None
    is_followed: bool = None


class Artists(MultiMediaBaseModel):
    def __init__(self, client: BaseClient, resp: Dict) -> None:
        self.users = [Artist(**artist) for artist in resp.pop("users")]
        super().__init__(client, **resp)



from typing import Dict, List
from pixivcat import BaseModel, MultiMediaBaseModel, BaseClient
from .user import Artist
from .illust import Illustration


class Preview(BaseModel):
    user: Artist
    illusts: List[Illustration]
    novels: list
    is_muted: bool

    def __init__(self, **kwds) -> None:
        self.illusts = [Illustration(**illust)
                        for illust in kwds.pop("illusts")]
        super().__init__(**kwds)


class Previews(MultiMediaBaseModel):
    def __init__(self, client: BaseClient, resp: Dict) -> None:
        self.previews = [Preview(**preview)
                         for preview in resp.pop("user_previews")]
        super().__init__(client=client, **resp)



from typing import List
from pixivcat import BaseModel, BaseClient, MultiMediaBaseModel


class BookmarkDetail(BaseModel):
    is_bookmarked: bool
    tags: List
    restrict: str


class Bookmarks(MultiMediaBaseModel):
    bookmark_tags: List

    def __init__(self, client: BaseClient, **kwds) -> None:
        super().__init__(client, **kwds)



from typing import Dict, List, Optional
from pixivcat import BaseModel, MultiMediaBaseModel
from pixivcat.user import Artist
from pixivcat.client import BaseClient


class Novel(BaseModel):
    id: int
    title: str
    caption: str
    restrict: int
    x_restrict: int
    is_original: bool
    image_urls: Dict
    create_date: str
    tags: List
    page_count: int
    text_length: int
    user: Artist
    series: Dict
    is_bookmarked: bool
    total_bookmarks: int
    total_view: int
    visible: bool
    total_comments: int
    is_muted: bool
    is_mypixiv_only: bool
    is_x_restricted: bool


class Novels(MultiMediaBaseModel):
    def __init__(self, client: BaseClient, resp: Dict) -> None:
        self.novels = [Novel(**novel) for novel in resp.pop("novels")]
        super().__init__(client=client, **resp)


class NovelSeriesDetail(BaseModel):
    id: int
    title: str
    caption: str
    is_original: bool
    is_concluded: bool
    content_count: int
    total_character_count: int
    user: Artist
    display_text: str


class NovelSeries(MultiMediaBaseModel):
    def __init__(self, client: BaseClient, resp: Dict) -> None:
        self.novel_series_detail = NovelSeriesDetail(
            **resp.pop("novel_series_detail"))
        self.novel_series_first_novel = Novel(
            **resp.pop("novel_series_first_novel"))
        self.novel_series_latest_novel = Novel(
            **resp.pop("novel_series_latest_novel"))
        self.novels = [Novel(**novel) for novel in resp.pop("novels")]
        super().__init__(client, **resp)



import contextvars
from typing import Dict, Union
from aiohttp import ClientSession
from aiohttp.client import ClientResponse
from multidict import CIMultiDict
session = contextvars.ContextVar("session")
main_loop = contextvars.ContextVar("loop")
PROXY = contextvars.ContextVar("proxy")


class Session(ClientSession):
    def __init__(self, proxy: str, *args, **kws) -> None:
        self.proxy = proxy
        PROXY.set(proxy)
        super().__init__(*args, **kws)

    async def _request(self, *args, **kws) -> ClientResponse:
        kws["proxy"] = self.proxy
        return await super()._request(*args, **kws)

    def update_headers(self, headers: Union[CIMultiDict, Dict, None] = None, **kws) -> None:
        """add some new headers"""
        self._default_headers.update(**headers or kws)



from typing import Dict


__version__ = "0.0.1"


class BaseClient:
    pass


class BaseModel:
    def __init__(self, **kwds) -> None:
        for _k, _v in self.__annotations__.items():
            val = kwds.get(_k, None)
            if not val:
                val = None
            elif not isinstance(val, _v):
                val = _v(**val) if isinstance(val, Dict) else _v(*val)
            setattr(self, _k, val)

    def __str__(self) -> str:
        return "<"+self.__class__.__name__+":\n"+"".join([f"\t| {_k}: {_v}\n" for _k, _v in self.__dict__.items() if _v])+"\n>"


class MultiMediaBaseModel:
    def __init__(self, client: BaseClient, **kwds) -> None:
        self.client = client
        self.next_url = kwds.pop("next_url", None)
        for k, v in kwds.items():
            setattr(self, k, v)

    async def next(self):
        if self.next_url:
            response = await self.client.request("GET", self.next_url)
            return self.__class__(client=self.client, resp=response)
        return None
